#!/usr/bin/env perl

use strict;
use Getopt::Long;
use Data::Dumper;

use DB_CONFIG;

my $species = 'Homo sapiens';
my $assembly = 'GRCh38';
my $phylo = "";
my $outdir = "/ensembl-data/$species";
my $verbose = '0';
my $help = '0';

####################################################################
## Parsing command line options	####################################
####################################################################
# USAGE: ./gene_extra_info.pl --species "Homo sapiens" --outdir ../../appl_db/ird_v1/hsa ...

## Parsing command line
GetOptions ('species=s' => \$species, 'assembly=s' => \$assembly, 'outdir=s' => \$outdir, 'phylo=s' => \$phylo,
			'ensembl-registry=s' => \$ENSEMBL_REGISTRY, 'ensembl-host=s' => \$ENSEMBL_HOST, 'ensembl-port=s' => \$ENSEMBL_PORT,
			'ensembl-user=s' => \$ENSEMBL_USER, 'verbose' => \$verbose, 'help' => \$help);

## Checking help parameter
print_usage() if $help;

## Printing parameters
print_parameters() if $verbose;

## Checking outdir parameter exist, otherwise create it
if (-d $outdir) {
	print "Writing files to directory '$outdir'...\n";
} else {
	print "Directory '$outdir' does not exist, creating directory...\n";
	mkdir $outdir or die "Couldn't create dir: [$outdir] ($!)";
}

####################################################################
## Ensembl API	####################################################
####################################################################
use Bio::EnsEMBL::DBSQL::DBAdaptor;
# use Bio::EnsEMBL::Funcgen::DBSQL::DBAdaptor;

## loading the registry with the adaptors
# print ("Loading $ENSEMBL_REGISTRY...\n");
# Bio::EnsEMBL::Registry->load_all("$ENSEMBL_REGISTRY");
# print ("done!\n");

if ($phylo eq "" || $phylo eq "vertebrate") {
	print ("In vertebrates section\n");
	if ($species eq "Homo sapiens" && $assembly eq "GRCh38") {
		print ("Human selected, assembly ".$assembly." selected, connecting to port ".$ENSEMBL_PORT."\n");
		Bio::EnsEMBL::Registry->load_registry_from_db(
			-host     => $ENSEMBL_HOST,
			-user     => $ENSEMBL_USER,
			-port     => $ENSEMBL_PORT,
			-verbose  => $verbose
		);
	} else {
		print ("Human selected, assembly ".$assembly." no supported\n");
	}
} else {
	print ("In no-vertebrates section\n");
	Bio::EnsEMBL::Registry->load_registry_from_db(
		-host => 'mysql-eg-publicsql.ebi.ac.uk',
		-port => 4157,
		-user => 'anonymous'
	);
}

print ("creating dbadaptors...\n");
my $gene_adaptor = Bio::EnsEMBL::Registry->get_adaptor($species, "core", "Gene");
# my $pba = Bio::EnsEMBL::Registry->get_adaptor($species, "funcgen", "ProbeSet");
####################################################################

## variables definition
my %dbnames = ();
my ($ngene, $ntrans) = (0, 0);
my ($gene, $translation, %uniq);

open (GENE_DESC, ">$outdir/description.txt") || die "Cannot open gene_description.txt file";
open (XREFS, ">$outdir/xrefs.txt") || die "Cannot open xref.txt file";

## for all the genes in the specie
foreach my $gene_id(@{$gene_adaptor->list_stable_ids()}) {
	$ngene++;
	$gene = $gene_adaptor->fetch_by_stable_id($gene_id);
	print $ngene."\t".$gene->stable_id."\t".$gene->description."\n" if $verbose;

	## Gene Info and sequence
	print GENE_DESC $gene->stable_id."\t".$gene->description."\n";

	## XREF TRANSCRIPTS
	foreach my $trans(@{$gene->get_all_Transcripts}) {
		$ntrans++;
		print "\t".$ntrans."\t".$trans->stable_id."\t".$trans->description."\n" if $verbose;

		# If trancript is a microrna we save it in xref
		if($trans->biotype eq 'miRNA') {
			print XREFS $trans->stable_id()."\t".$gene->external_name()."\t".&get_dbname_short("miRNA gene")."\t"."miRNA gene"."\t".$gene->description()."\n";
		}

		# Reference to ensembl gene id
		print XREFS $trans->stable_id()."\t".$gene->stable_id()."\t".&get_dbname_short("Ensembl gene")."\t"."Ensembl gene"."\t".$gene->description()."\n";

		# Additional synonyms
		my ($display_xref_entry);
		$display_xref_entry = $gene->display_xref();
		if (defined($display_xref_entry)) {
			foreach my $synonym_str(@{ $display_xref_entry->get_all_synonyms }) {
				print XREFS $trans->stable_id() . "\t" . $synonym_str . "\t" . &get_dbname_short("Ensembl gene") . "\t" . "Ensembl gene" . "\t" . $gene->description() . "\n";
			}
		}

		#			print XREF $ntrans."\t".$gene->external_name."\t".&get_dbname($gene->get_all_DBEntries->[0]->{'db_display_name'})."\t".$gene->get_all_DBEntries->[0]->{'description'}."\n";
		#			print XREFS $trans->stable_id()."\t".$gene->external_name()."\t".&get_dbname_short($gene->get_all_DBEntries->[0]->{'db_display_name'})."\t".$gene->get_all_DBEntries->[0]->{'db_display_name'}."\t".$gene->get_all_DBEntries->[0]->{'description'}."\n";
		foreach my $dbentry(@{$gene->get_all_DBEntries}){
			print XREFS $trans->stable_id()."\t".$dbentry->{'display_id'}."\t".&get_dbname_short($dbentry->{'db_display_name'})."\t".$dbentry->{'db_display_name'}."\t".$dbentry->{'description'}."\n";
		}

		# Reference to ensembl transcript id
		print XREFS $trans->stable_id()."\t".$trans->stable_id()."\t".&get_dbname_short("Ensembl transcript")."\t"."Ensembl transcript"."\t".$trans->description()."\n";

		##### Translation ######################################################
		# Reference to ensembl prot id
		$translation = $trans->translation();
		if (defined($translation)) {
			print XREFS $trans->stable_id()."\t".$translation->stable_id."\t".&get_dbname_short("Ensembl protein")."\t"."Ensembl protein"."\t".$trans->description()."\n";
			# foreach my $dblink(@{ $translation->get_all_xrefs }) {
			# 	print XREFS $trans->stable_id()."\t".$dblink->display_id()."\t".&get_dbname_short($dblink->db_display_name())."\t".$dblink->db_display_name()."\t".$dblink->description()."\n";
			# }

			%uniq = ();
			foreach my $df(@{$translation->get_all_ProteinFeatures()}){
				if( $df->interpro_ac() ne "" && not defined($uniq{$df->interpro_ac()})) {
					print XREFS $trans->stable_id()."\t".$df->interpro_ac()."\t".&get_dbname_short("InterPro")."\t"."InterPro"."\t".$df->idesc()."\n";
					$uniq{$df->interpro_ac()} = 1;
				}
			}
		}

		foreach my $dblink(@{ $trans->get_all_xrefs }) {
			print XREFS $trans->stable_id()."\t".$dblink->display_id()."\t".&get_dbname_short($dblink->db_display_name())."\t".$dblink->db_display_name()."\t".$dblink->description()."\n";

			if ($dblink->display_id() =~ /(\w+)\.\d$/){
				print XREFS $trans->stable_id()."\t".$1."\t".&get_dbname_short($dblink->db_display_name())."\t".$dblink->db_display_name()."\t".$dblink->description()."\n";
			}
			if ($dblink->display_id() =~ /(\w+)\-\d$/){
				print XREFS $trans->stable_id()."\t".$1."\t".&get_dbname_short($dblink->db_display_name())."\t".$dblink->db_display_name()."\t".$dblink->description()."\n";
			}
			if (defined ($trembl_species{$species}) && $dblink->display_id() =~ /(\w+)_$trembl_species{$species}/){
				print XREFS $trans->stable_id()."\t".$1."\t".&get_dbname_short($dblink->db_display_name())."\t".$dblink->db_display_name()."\t".$dblink->description()."\n";
			}
		}

		## MICROARRYAYS - DEPRECATED!
		# if (defined $pba) {
		# 	my @probesets = @{$pba->fetch_all_by_transcript_stable_id($trans->stable_id)};
		# 	foreach my $probeset (@probesets){
		#
		# 		my @arrays_string = map $_->name, @{$probeset->get_all_Arrays};
		# 		my $dbe_info;
		#
		# 		#Now get linkage_annotation
		# 		#              foreach my $dbe(@{$probeset->get_all_Transcript_DBEntries($trans)}){
		# 		#This will return all ProbeSet DBEntries for this transcript
		# 		#There should really be one max per transcript per probeset/probe
		# 		#            	$dbe_info = $dbe->linkage_annotation;
		# 		#              }
		#
		# 		#              print "\t".$probeset->name." on arrays ".$arrays_string." with Probe hits $dbe_info\n";
		# 		foreach my $array_string(@arrays_string) {
		# 			print XREFS $trans->stable_id()."\t".$probeset->name."\t".&get_dbname_short($array_string)."\t".$array_string."\t\n";
		# 		}
		# 	}
		# }
	}
}

close(GENE_DESC);
close(XREFS);


sub get_dbname_short {
	my $display_name = shift;
	my $dbname = $display_name;
	if ($display_name =~ /Affymx Microarray/) {
		$dbname =~ s/Affymx Microarray/affy/;
	}elsif ($display_name =~ /MIM/) {
		$dbname =~ s/MIM/omim/;
	}elsif ($display_name =~ /Codelink/) {
		$dbname = "codelink";
	}elsif ($display_name =~ /Swiss-Prot/) {
		$dbname =~ s/Swiss-Prot/swissprot/;
	}elsif ($display_name =~ /MEROPS/) {
		$dbname =~ s/MEROPS - the Peptidase Database/merops/;
	}
	$dbname =~ s/ /_/g;
	$dbname =~ s/-/_/g;
	$dbname =~ s/\(//g;
	$dbname =~ s/\)//g;
	$dbnames{$display_name} = lc($dbname);
	return lc($dbname);
}

sub print_parameters {
	print "Parameters: ";
	print "species: $species, outdir: $outdir, ";
	print "ensembl-registry: $ENSEMBL_REGISTRY, ";
	print "ensembl-host: $ENSEMBL_HOST, ensembl-port: $ENSEMBL_PORT, ";
	print "ensembl-user: $ENSEMBL_USER, verbose: $verbose, help: $help";
	print "\n";
}

sub print_usage {
	print "Usage:   $0 [--species] [--outdir] [--ensembl-host] [--ensembl-port] [--ensembl-user] [--verbose] [--help]\n";
	print "\tDescription:\n";
	print "\t species: scientific name, i.e.: Homo sapiens\n";
	print "\t outdir: default directory /tmp/species\n";
	print "\t ensembl-host: Ensembl database host\n";
	print "\t ensembl-port: Ensembl database port\n";
	print "\t ensembl-user: Ensembl database user\n";
	print "\t verbose: print logs\n";
	print "\t help: print this help\n";
	print "\n";
	exit(0);
}
